@@ -27,7 +27,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: ipc.c,v 1.8 2004/09/04 03:39:20 roland Exp $
+ *	$Id: ipc.c,v 1.17 2007/01/11 23:19:55 ldv Exp $
  */
 
 #include "defs.h"
@@ -67,7 +67,6 @@
 # define IPC_64 0x100
 #endif
 
-extern const struct xlat openmodes[];
 extern void printsigevent(struct tcb *tcp, long arg);
 
 static const struct xlat msgctl_flags[] = {
@@ -123,6 +122,12 @@ static const struct xlat resource_flags[] = {
 	{ IPC_CREAT,	"IPC_CREAT"	},
 	{ IPC_EXCL,	"IPC_EXCL"	},
 	{ IPC_NOWAIT,	"IPC_NOWAIT"	},
+	{ 0,		NULL		},
+};
+
+static const struct xlat shm_resource_flags[] = {
+	{ IPC_CREAT,	"IPC_CREAT"	},
+	{ IPC_EXCL,	"IPC_EXCL"	},
 #ifdef SHM_HUGETLB
 	{ SHM_HUGETLB,	"SHM_HUGETLB"	},
 #endif
@@ -152,11 +157,11 @@ struct tcb *tcp;
 {
 	if (entering(tcp)) {
 		if (tcp->u_arg[0])
-			tprintf("%lu", tcp->u_arg[0]);
+			tprintf("%#lx", tcp->u_arg[0]);
 		else
 			tprintf("IPC_PRIVATE");
 		tprintf(", ");
-		if (printflags(resource_flags, tcp->u_arg[1] & ~0777) != 0)
+		if (printflags(resource_flags, tcp->u_arg[1] & ~0777, NULL) != 0)
 			tprintf("|");
 		tprintf("%#lo", tcp->u_arg[1] & 0777);
 	}
@@ -171,17 +176,31 @@ struct tcb *tcp;
 # define PRINTCTL printxval
 #endif
 
+static int
+indirect_ipccall(tcp)
+struct tcb *tcp;
+{
+#ifdef LINUX
+#ifdef X86_64
+	return current_personality > 0;
+#endif
+#if defined IA64
+	return tcp->scno < 1024; /* ia32 emulation syscalls are low */
+#endif
+#if !defined MIPS && !defined HPPA
+	return 1;
+#endif
+#endif	/* LINUX */
+	return 0;
+}
+
 int sys_msgctl(tcp)
 struct tcb *tcp;
 {
 	if (entering(tcp)) {
 		tprintf("%lu, ", tcp->u_arg[0]);
 		PRINTCTL(msgctl_flags, tcp->u_arg[1], "MSG_???");
-#ifdef LINUX
-		tprintf(", %#lx", tcp->u_arg[3]);
-#else /* !LINUX */
-		tprintf(", %#lx", tcp->u_arg[2]);
-#endif /* !LINUX */
+		tprintf(", %#lx", tcp->u_arg[indirect_ipccall(tcp) ? 3 : 2]);
 	}
 	return 0;
 }
@@ -193,25 +212,23 @@ struct tcb *tcp;
 
 	if (entering(tcp)) {
 		tprintf("%lu", tcp->u_arg[0]);
-#ifdef LINUX
-		umove(tcp, tcp->u_arg[3], &mtype);
-		tprintf(", {%lu, ", mtype);
-		printstr(tcp, tcp->u_arg[3] + sizeof(long),
-			tcp->u_arg[1]);
-		tprintf("}, %lu", tcp->u_arg[1]);
-		tprintf(", ");
-		if (printflags(msg_flags, tcp->u_arg[2]) == 0)
-			tprintf("0");
-#else /* !LINUX */
-		umove(tcp, tcp->u_arg[1], &mtype);
-		tprintf(", {%lu, ", mtype);
-		printstr(tcp, tcp->u_arg[1] + sizeof(long),
-			tcp->u_arg[2]);
-		tprintf("}, %lu", tcp->u_arg[2]);
-		tprintf(", ");
-		if (printflags(msg_flags, tcp->u_arg[3]) == 0)
-			tprintf("0");
-#endif /* !LINUX */
+		if (indirect_ipccall(tcp)) {
+			umove(tcp, tcp->u_arg[3], &mtype);
+			tprintf(", {%lu, ", mtype);
+			printstr(tcp, tcp->u_arg[3] + sizeof(long),
+				 tcp->u_arg[1]);
+			tprintf("}, %lu", tcp->u_arg[1]);
+			tprintf(", ");
+			printflags(msg_flags, tcp->u_arg[2], "MSG_???");
+		} else {
+			umove(tcp, tcp->u_arg[1], &mtype);
+			tprintf(", {%lu, ", mtype);
+			printstr(tcp, tcp->u_arg[1] + sizeof(long),
+				 tcp->u_arg[2]);
+			tprintf("}, %lu", tcp->u_arg[2]);
+			tprintf(", ");
+			printflags(msg_flags, tcp->u_arg[3], "MSG_???");
+		}
 	}
 	return 0;
 }
@@ -220,38 +237,35 @@ int sys_msgrcv(tcp)
 struct tcb *tcp;
 {
 	long mtype;
-#ifdef LINUX
-	struct ipc_wrapper {
-		struct msgbuf *msgp;
-		long msgtyp;
-	} tmp;
-#endif
 
-
-	if (exiting(tcp)) {
+	if (entering(tcp)) {
+		tprintf("%lu, ", tcp->u_arg[0]);
+	} else {
 		tprintf("%lu", tcp->u_arg[0]);
-#ifdef LINUX
-		umove(tcp, tcp->u_arg[3], &tmp);
-		umove(tcp, (long) tmp.msgp, &mtype);
-		tprintf(", {%lu, ", mtype);
-		printstr(tcp, (long) (tmp.msgp) + sizeof(long),
-			tcp->u_arg[1]);
-		tprintf("}, %lu", tcp->u_arg[1]);
-		tprintf(", %ld", tmp.msgtyp);
-		tprintf(", ");
-		if (printflags(msg_flags, tcp->u_arg[2]) == 0)
-			tprintf("0");
-#else /* !LINUX */
-		umove(tcp, tcp->u_arg[1], &mtype);
-		tprintf(", {%lu, ", mtype);
-		printstr(tcp, tcp->u_arg[1] + sizeof(long),
-			tcp->u_arg[2]);
-		tprintf("}, %lu", tcp->u_arg[2]);
-		tprintf(", %ld", tcp->u_arg[3]);
-		tprintf(", ");
-		if (printflags(msg_flags, tcp->u_arg[4]) == 0)
-			tprintf("0");
-#endif /* !LINUX */
+		if (indirect_ipccall(tcp)) {
+			struct ipc_wrapper {
+				struct msgbuf *msgp;
+				long msgtyp;
+			} tmp;
+			umove(tcp, tcp->u_arg[3], &tmp);
+			umove(tcp, (long) tmp.msgp, &mtype);
+			tprintf(", {%lu, ", mtype);
+			printstr(tcp, (long) (tmp.msgp) + sizeof(long),
+				 tcp->u_arg[1]);
+			tprintf("}, %lu", tcp->u_arg[1]);
+			tprintf(", %ld", tmp.msgtyp);
+			tprintf(", ");
+			printflags(msg_flags, tcp->u_arg[2], "MSG_???");
+		} else {
+			umove(tcp, tcp->u_arg[1], &mtype);
+			tprintf("{%lu, ", mtype);
+			printstr(tcp, tcp->u_arg[1] + sizeof(long),
+				 tcp->u_arg[2]);
+			tprintf("}, %lu", tcp->u_arg[2]);
+			tprintf(", %ld", tcp->u_arg[3]);
+			tprintf(", ");
+			printflags(msg_flags, tcp->u_arg[4], "MSG_???");
+		}
 	}
 	return 0;
 }
@@ -261,13 +275,13 @@ struct tcb *tcp;
 {
 	if (entering(tcp)) {
 		tprintf("%lu", tcp->u_arg[0]);
-#ifdef LINUX
-		tprintf(", %#lx", tcp->u_arg[3]);
-		tprintf(", %lu", tcp->u_arg[1]);
-#else /* !LINUX */
-		tprintf(", %#lx", tcp->u_arg[1]);
-		tprintf(", %lu", tcp->u_arg[2]);
-#endif /* !LINUX */
+		if (indirect_ipccall(tcp)) {
+			tprintf(", %#lx", tcp->u_arg[3]);
+			tprintf(", %lu", tcp->u_arg[1]);
+		} else {
+			tprintf(", %#lx", tcp->u_arg[1]);
+			tprintf(", %lu", tcp->u_arg[2]);
+		}
 	}
 	return 0;
 }
@@ -278,9 +292,15 @@ struct tcb *tcp;
 {
 	if (entering(tcp)) {
 		tprintf("%lu", tcp->u_arg[0]);
-		tprintf(", %#lx", tcp->u_arg[3]);
-		tprintf(", %lu, ", tcp->u_arg[1]);
-		printtv(tcp, tcp->u_arg[5]);
+		if (indirect_ipccall(tcp)) {
+			tprintf(", %#lx", tcp->u_arg[3]);
+			tprintf(", %lu, ", tcp->u_arg[1]);
+			printtv(tcp, tcp->u_arg[5]);
+		} else {
+			tprintf(", %#lx", tcp->u_arg[1]);
+			tprintf(", %lu, ", tcp->u_arg[2]);
+			printtv(tcp, tcp->u_arg[3]);
+		}
 	}
 	return 0;
 }
@@ -291,12 +311,12 @@ struct tcb *tcp;
 {
 	if (entering(tcp)) {
 		if (tcp->u_arg[0])
-			tprintf("%lu", tcp->u_arg[0]);
+			tprintf("%#lx", tcp->u_arg[0]);
 		else
 			tprintf("IPC_PRIVATE");
 		tprintf(", %lu", tcp->u_arg[1]);
 		tprintf(", ");
-		if (printflags(resource_flags, tcp->u_arg[2] & ~0777) != 0)
+		if (printflags(resource_flags, tcp->u_arg[2] & ~0777, NULL) != 0)
 			tprintf("|");
 		tprintf("%#lo", tcp->u_arg[2] & 0777);
 	}
@@ -320,12 +340,12 @@ struct tcb *tcp;
 {
 	if (entering(tcp)) {
 		if (tcp->u_arg[0])
-			tprintf("%lu", tcp->u_arg[0]);
+			tprintf("%#lx", tcp->u_arg[0]);
 		else
 			tprintf("IPC_PRIVATE");
 		tprintf(", %lu", tcp->u_arg[1]);
 		tprintf(", ");
-		if (printflags(resource_flags, tcp->u_arg[2] & ~0777) != 0)
+		if (printflags(shm_resource_flags, tcp->u_arg[2] & ~0777, NULL) != 0)
 			tprintf("|");
 		tprintf("%#lo", tcp->u_arg[2] & 0777);
 	}
@@ -338,11 +358,11 @@ struct tcb *tcp;
 	if (entering(tcp)) {
 		tprintf("%lu, ", tcp->u_arg[0]);
 		PRINTCTL(shmctl_flags, tcp->u_arg[1], "SHM_???");
-#ifdef LINUX
-		tprintf(", %#lx", tcp->u_arg[3]);
-#else /* !LINUX */
-		tprintf(", %#lx", tcp->u_arg[2]);
-#endif /* !LINUX */
+		if (indirect_ipccall(tcp)) {
+			tprintf(", %#lx", tcp->u_arg[3]);
+		} else {
+			tprintf(", %#lx", tcp->u_arg[2]);
+		}
 	}
 	return 0;
 }
@@ -356,17 +376,15 @@ struct tcb *tcp;
 
 	if (exiting(tcp)) {
 		tprintf("%lu", tcp->u_arg[0]);
-#ifdef LINUX
-		tprintf(", %#lx", tcp->u_arg[3]);
-		tprintf(", ");
-		if (printflags(shm_flags, tcp->u_arg[1]) == 0)
-			tprintf("0");
-#else /* !LINUX */
-		tprintf(", %#lx", tcp->u_arg[1]);
-		tprintf(", ");
-		if (printflags(shm_flags, tcp->u_arg[2]) == 0)
-			tprintf("0");
-#endif /* !LINUX */
+		if (indirect_ipccall(tcp)) {
+			tprintf(", %#lx", tcp->u_arg[3]);
+			tprintf(", ");
+			printflags(shm_flags, tcp->u_arg[1], "SHM_???");
+		} else {
+			tprintf(", %#lx", tcp->u_arg[1]);
+			tprintf(", ");
+			printflags(shm_flags, tcp->u_arg[2], "SHM_???");
+		}
 		if (syserror(tcp))
 			return 0;
 #ifdef LINUX
@@ -382,26 +400,27 @@ struct tcb *tcp;
 int sys_shmdt(tcp)
 struct tcb *tcp;
 {
-	if (entering(tcp))
-#ifdef LINUX
-		tprintf("%#lx", tcp->u_arg[3]);
-#else /* !LINUX */
-		tprintf("%#lx", tcp->u_arg[0]);
-#endif /* !LINUX */
+	if (entering(tcp)) {
+		if (indirect_ipccall(tcp)) {
+			tprintf("%#lx", tcp->u_arg[3]);
+		} else {
+			tprintf("%#lx", tcp->u_arg[0]);
+		}
+	}
 	return 0;
 }
 
 #endif /* defined(LINUX) || defined(SUNOS4) || defined(FREEBSD) */
 
 #ifdef LINUX
-int sys_mq_open(tcp)
-struct tcb *tcp;
+int
+sys_mq_open(struct tcb *tcp)
 {
 	if (entering(tcp)) {
 		printpath(tcp, tcp->u_arg[0]);
 		tprintf(", ");
 		/* flags */
-		printflags(openmodes, tcp->u_arg[1] + 1);
+		tprint_open_modes(tcp, tcp->u_arg[1]);
 		if (tcp->u_arg[1] & O_CREAT) {
 # ifndef HAVE_MQUEUE_H
 			tprintf(", %lx", tcp->u_arg[2]);
@@ -420,8 +439,8 @@ struct tcb *tcp;
 	return 0;
 }
 
-int sys_mq_timedsend(tcp)
-struct tcb *tcp;
+int
+sys_mq_timedsend(struct tcb *tcp)
 {
 	if (entering(tcp)) {
 		tprintf("%ld, ", tcp->u_arg[0]);
@@ -432,8 +451,8 @@ struct tcb *tcp;
 	return 0;
 }
 
-int sys_mq_timedreceive(tcp)
-struct tcb *tcp;
+int
+sys_mq_timedreceive(struct tcb *tcp)
 {
 	if (entering(tcp))
 		tprintf("%ld, ", tcp->u_arg[0]);
@@ -445,8 +464,8 @@ struct tcb *tcp;
 	return 0;
 }
 
-int sys_mq_notify(tcp)
-struct tcb *tcp;
+int
+sys_mq_notify(struct tcb *tcp)
 {
 	if (entering(tcp)) {
 		tprintf("%ld, ", tcp->u_arg[0]);
@@ -455,9 +474,8 @@ struct tcb *tcp;
 	return 0;
 }
 
-static void printmqattr(tcp, addr)
-struct tcb *tcp;
-long addr;
+static void
+printmqattr(struct tcb *tcp, long addr)
 {
 	if (addr == 0)
 		tprintf("NULL");
@@ -471,15 +489,15 @@ long addr;
 			return;
 		}
 		tprintf("{mq_flags=");
-		printflags(openmodes, attr.mq_flags + 1);
+		tprint_open_modes(tcp, attr.mq_flags);
 		tprintf(", mq_maxmsg=%ld, mq_msgsize=%ld, mq_curmsg=%ld}",
 			attr.mq_maxmsg, attr.mq_msgsize, attr.mq_curmsgs);
 # endif
 	}
 }
 
-int sys_mq_getsetattr(tcp)
-struct tcb *tcp;
+int
+sys_mq_getsetattr(struct tcb *tcp)
 {
 	if (entering(tcp)) {
 		tprintf("%ld, ", tcp->u_arg[0]);
