@@ -27,7 +27,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: strace.c,v 1.68 2006/01/12 09:50:49 roland Exp $
+ *	$Id: strace.c,v 1.88 2008/08/06 21:38:52 kratochvil Exp $
  */
 
 #include "defs.h"
@@ -46,6 +46,21 @@
 #include <limits.h>
 #include <dirent.h>
 
+#ifdef LINUX
+# include <asm/unistd.h>
+# if defined __NR_tgkill
+#  define my_tgkill(pid, tid, sig) syscall (__NR_tgkill, (pid), (tid), (sig))
+# elif defined __NR_tkill
+#  define my_tgkill(pid, tid, sig) syscall (__NR_tkill, (tid), (sig))
+# else
+   /* kill() may choose arbitrarily the target task of the process group
+      while we later wait on a that specific TID.  PID process waits become
+      TID task specific waits for a process under ptrace(2).  */
+#  warning "Neither tkill(2) nor tgkill(2) available, risk of strace hangs!"
+#  define my_tgkill(pid, tid, sig) kill ((tid), (sig))
+# endif
+#endif
+
 #if defined(IA64) && defined(LINUX)
 # include <asm/ptrace_offsets.h>
 #endif
@@ -63,27 +78,27 @@
 #endif
 #endif
 
-int debug = 0, followfork = 0, followvfork = 0, interactive = 0;
-int rflag = 0, tflag = 0, dtime = 0, cflag = 0;
-int iflag = 0, xflag = 0, qflag = 0;
-int pflag_seen = 0;
+int debug = 0, followfork = 0;
+int dtime = 0, cflag = 0, xflag = 0, qflag = 0;
+static int iflag = 0, interactive = 0, pflag_seen = 0, rflag = 0, tflag = 0;
 
 /* Sometimes we want to print only succeeding syscalls. */
 int not_failing_only = 0;
 
-char *username = NULL;
+static char *username = NULL;
 uid_t run_uid;
 gid_t run_gid;
 
 int acolumn = DEFAULT_ACOLUMN;
 int max_strlen = DEFAULT_STRLEN;
-char *outfname = NULL;
+static char *outfname = NULL;
 FILE *outf;
 struct tcb **tcbtab;
 unsigned int nprocs, tcbtabsize;
 char *progname;
 extern char **environ;
 
+static int detach P((struct tcb *tcp, int sig));
 static int trace P((void));
 static void cleanup P((void));
 static void interrupt P((int sig));
@@ -176,27 +191,457 @@ foobar()
 #endif /* MIPS */
 #endif /* SVR4 */
 
+static int
+set_cloexec_flag(int fd)
+{
+	int     flags, newflags;
+
+	if ((flags = fcntl(fd, F_GETFD, 0)) < 0)
+	{
+		fprintf(stderr, "%s: fcntl F_GETFD: %s\n",
+			progname, strerror(errno));
+		return -1;
+	}
+
+	newflags = flags | FD_CLOEXEC;
+	if (flags == newflags)
+		return 0;
+
+	if (fcntl(fd, F_SETFD, newflags) < 0)
+	{
+		fprintf(stderr, "%s: fcntl F_SETFD: %s\n",
+			progname, strerror(errno));
+		return -1;
+	}
+
+	return 0;
+}
+
+/*
+ * When strace is setuid executable, we have to swap uids
+ * before and after filesystem and process management operations.
+ */
+static void
+swap_uid(void)
+{
+#ifndef SVR4
+	int euid = geteuid(), uid = getuid();
+
+	if (euid != uid && setreuid(euid, uid) < 0)
+	{
+		fprintf(stderr, "%s: setreuid: %s\n",
+			progname, strerror(errno));
+		exit(1);
+	}
+#endif
+}
+
+#if _LFS64_LARGEFILE
+# define fopen_for_output fopen64
+#else
+# define fopen_for_output fopen
+#endif
+
+static FILE *
+strace_fopen(const char *path, const char *mode)
+{
+	FILE *fp;
+
+	swap_uid();
+	if ((fp = fopen_for_output(path, mode)) == NULL)
+		fprintf(stderr, "%s: can't fopen '%s': %s\n",
+			progname, path, strerror(errno));
+	swap_uid();
+	if (fp && set_cloexec_flag(fileno(fp)) < 0)
+	{
+		fclose(fp);
+		fp = NULL;
+	}
+	return fp;
+}
+
+static int popen_pid = -1;
+
+#ifndef _PATH_BSHELL
+# define _PATH_BSHELL "/bin/sh"
+#endif
+
+/*
+ * We cannot use standard popen(3) here because we have to distinguish
+ * popen child process from other processes we trace, and standard popen(3)
+ * does not export its child's pid.
+ */
+static FILE *
+strace_popen(const char *command)
+{
+	int     fds[2];
+
+	swap_uid();
+	if (pipe(fds) < 0)
+	{
+		fprintf(stderr, "%s: pipe: %s\n",
+			progname, strerror(errno));
+		swap_uid();
+		return NULL;
+	}
+
+	if (set_cloexec_flag(fds[1]) < 0)
+	{
+		close(fds[0]);
+		close(fds[1]);
+		swap_uid();
+		return NULL;
+	}
+
+	if ((popen_pid = fork()) == -1)
+	{
+		fprintf(stderr, "%s: fork: %s\n",
+			progname, strerror(errno));
+		close(fds[0]);
+		close(fds[1]);
+		swap_uid();
+		return NULL;
+	}
+
+	if (popen_pid)
+	{
+		/* parent */
+		close(fds[0]);
+		swap_uid();
+		return fdopen(fds[1], "w");
+	} else
+	{
+		/* child */
+		close(fds[1]);
+		if (fds[0] && (dup2(fds[0], 0) || close(fds[0])))
+		{
+			fprintf(stderr, "%s: dup2: %s\n",
+				progname, strerror(errno));
+			_exit(1);
+		}
+		execl(_PATH_BSHELL, "sh", "-c", command, NULL);
+		fprintf(stderr, "%s: execl: %s: %s\n",
+			progname, _PATH_BSHELL, strerror(errno));
+		_exit(1);
+	}
+}
+
+static int
+newoutf(struct tcb *tcp)
+{
+	if (outfname && followfork > 1) {
+		char name[MAXPATHLEN];
+		FILE *fp;
+
+		sprintf(name, "%s.%u", outfname, tcp->pid);
+		if ((fp = strace_fopen(name, "w")) == NULL)
+			return -1;
+		tcp->outf = fp;
+	}
+	return 0;
+}
+
+static void
+startup_attach(void)
+{
+	int tcbi;
+	struct tcb *tcp;
+
+	/*
+	 * Block user interruptions as we would leave the traced
+	 * process stopped (process state T) if we would terminate in
+	 * between PTRACE_ATTACH and wait4 () on SIGSTOP.
+	 * We rely on cleanup () from this point on.
+	 */
+	if (interactive)
+		sigprocmask(SIG_BLOCK, &blocked_set, NULL);
+
+	for (tcbi = 0; tcbi < tcbtabsize; tcbi++) {
+		tcp = tcbtab[tcbi];
+		if (!(tcp->flags & TCB_INUSE) || !(tcp->flags & TCB_ATTACHED))
+			continue;
+#ifdef LINUX
+		if (tcp->flags & TCB_CLONE_THREAD)
+			continue;
+#endif
+		/* Reinitialize the output since it may have changed. */
+		tcp->outf = outf;
+		if (newoutf(tcp) < 0)
+			exit(1);
+
+#ifdef USE_PROCFS
+		if (proc_open(tcp, 1) < 0) {
+			fprintf(stderr, "trouble opening proc file\n");
+			droptcb(tcp);
+			continue;
+		}
+#else /* !USE_PROCFS */
+# ifdef LINUX
+		if (followfork) {
+			char procdir[MAXPATHLEN];
+			DIR *dir;
+
+			sprintf(procdir, "/proc/%d/task", tcp->pid);
+			dir = opendir(procdir);
+			if (dir != NULL) {
+				unsigned int ntid = 0, nerr = 0;
+				struct dirent *de;
+				int tid;
+				while ((de = readdir(dir)) != NULL) {
+					if (de->d_fileno == 0 ||
+					    de->d_name[0] == '.')
+						continue;
+					tid = atoi(de->d_name);
+					if (tid <= 0)
+						continue;
+					++ntid;
+					if (ptrace(PTRACE_ATTACH, tid,
+						   (char *) 1, 0) < 0)
+						++nerr;
+					else if (tid != tcbtab[tcbi]->pid) {
+						if (nprocs == tcbtabsize &&
+						    expand_tcbtab())
+							tcp = NULL;
+						else
+							tcp = alloctcb(tid);
+						if (tcp == NULL)
+							exit(1);
+						tcp->flags |= TCB_ATTACHED|TCB_CLONE_THREAD|TCB_CLONE_DETACHED|TCB_FOLLOWFORK;
+						tcbtab[tcbi]->nchildren++;
+						tcbtab[tcbi]->nclone_threads++;
+						tcbtab[tcbi]->nclone_detached++;
+						tcp->parent = tcbtab[tcbi];
+					}
+					if (interactive) {
+						sigprocmask(SIG_SETMASK, &empty_set, NULL);
+						if (interrupted)
+							return;
+						sigprocmask(SIG_BLOCK, &blocked_set, NULL);
+					}
+				}
+				closedir(dir);
+				if (nerr == ntid) {
+					perror("attach: ptrace(PTRACE_ATTACH, ...)");
+					droptcb(tcp);
+					continue;
+				}
+				if (!qflag) {
+					ntid -= nerr;
+					if (ntid > 1)
+						fprintf(stderr, "\
+Process %u attached with %u threads - interrupt to quit\n",
+							tcp->pid, ntid);
+					else
+						fprintf(stderr, "\
+Process %u attached - interrupt to quit\n",
+							tcp->pid);
+				}
+				continue;
+			}
+		}
+# endif
+		if (ptrace(PTRACE_ATTACH, tcp->pid, (char *) 1, 0) < 0) {
+			perror("attach: ptrace(PTRACE_ATTACH, ...)");
+			droptcb(tcp);
+			continue;
+		}
+		/* INTERRUPTED is going to be checked at the top of TRACE.  */
+#endif /* !USE_PROCFS */
+		if (!qflag)
+			fprintf(stderr,
+				"Process %u attached - interrupt to quit\n",
+				tcp->pid);
+	}
+
+	if (interactive)
+		sigprocmask(SIG_SETMASK, &empty_set, NULL);
+}
+
+static void
+startup_child (char **argv)
+{
+	struct stat statbuf;
+	const char *filename;
+	char pathname[MAXPATHLEN];
+	int pid = 0;
+	struct tcb *tcp;
+
+	filename = argv[0];
+	if (strchr(filename, '/')) {
+		if (strlen(filename) > sizeof pathname - 1) {
+			errno = ENAMETOOLONG;
+			perror("strace: exec");
+			exit(1);
+		}
+		strcpy(pathname, filename);
+	}
+#ifdef USE_DEBUGGING_EXEC
+	/*
+	 * Debuggers customarily check the current directory
+	 * first regardless of the path but doing that gives
+	 * security geeks a panic attack.
+	 */
+	else if (stat(filename, &statbuf) == 0)
+		strcpy(pathname, filename);
+#endif /* USE_DEBUGGING_EXEC */
+	else {
+		char *path;
+		int m, n, len;
+
+		for (path = getenv("PATH"); path && *path; path += m) {
+			if (strchr(path, ':')) {
+				n = strchr(path, ':') - path;
+				m = n + 1;
+			}
+			else
+				m = n = strlen(path);
+			if (n == 0) {
+				if (!getcwd(pathname, MAXPATHLEN))
+					continue;
+				len = strlen(pathname);
+			}
+			else if (n > sizeof pathname - 1)
+				continue;
+			else {
+				strncpy(pathname, path, n);
+				len = n;
+			}
+			if (len && pathname[len - 1] != '/')
+				pathname[len++] = '/';
+			strcpy(pathname + len, filename);
+			if (stat(pathname, &statbuf) == 0 &&
+			    /* Accept only regular files
+			       with some execute bits set.
+			       XXX not perfect, might still fail */
+			    S_ISREG(statbuf.st_mode) &&
+			    (statbuf.st_mode & 0111))
+				break;
+		}
+	}
+	if (stat(pathname, &statbuf) < 0) {
+		fprintf(stderr, "%s: %s: command not found\n",
+			progname, filename);
+		exit(1);
+	}
+	switch (pid = fork()) {
+	case -1:
+		perror("strace: fork");
+		cleanup();
+		exit(1);
+		break;
+	case 0: {
+#ifdef USE_PROCFS
+		if (outf != stderr) close (fileno (outf));
+#ifdef MIPS
+		/* Kludge for SGI, see proc_open for details. */
+		sa.sa_handler = foobar;
+		sa.sa_flags = 0;
+		sigemptyset(&sa.sa_mask);
+		sigaction(SIGINT, &sa, NULL);
+#endif /* MIPS */
+#ifndef FREEBSD
+		pause();
+#else /* FREEBSD */
+		kill(getpid(), SIGSTOP); /* stop HERE */
+#endif /* FREEBSD */
+#else /* !USE_PROCFS */
+		if (outf!=stderr)
+			close(fileno (outf));
+
+		if (ptrace(PTRACE_TRACEME, 0, (char *) 1, 0) < 0) {
+			perror("strace: ptrace(PTRACE_TRACEME, ...)");
+			exit(1);
+		}
+		if (debug)
+			kill(getpid(), SIGSTOP);
+
+		if (username != NULL || geteuid() == 0) {
+			uid_t run_euid = run_uid;
+			gid_t run_egid = run_gid;
+
+			if (statbuf.st_mode & S_ISUID)
+				run_euid = statbuf.st_uid;
+			if (statbuf.st_mode & S_ISGID)
+				run_egid = statbuf.st_gid;
+
+			/*
+			 * It is important to set groups before we
+			 * lose privileges on setuid.
+			 */
+			if (username != NULL) {
+				if (initgroups(username, run_gid) < 0) {
+					perror("initgroups");
+					exit(1);
+				}
+				if (setregid(run_gid, run_egid) < 0) {
+					perror("setregid");
+					exit(1);
+				}
+				if (setreuid(run_uid, run_euid) < 0) {
+					perror("setreuid");
+					exit(1);
+				}
+			}
+		}
+		else
+			setreuid(run_uid, run_uid);
+
+		/*
+		 * Induce an immediate stop so that the parent
+		 * will resume us with PTRACE_SYSCALL and display
+		 * this execve call normally.
+		 */
+		kill(getpid(), SIGSTOP);
+#endif /* !USE_PROCFS */
+
+		execv(pathname, argv);
+		perror("strace: exec");
+		_exit(1);
+		break;
+	}
+	default:
+		if ((tcp = alloctcb(pid)) == NULL) {
+			cleanup();
+			exit(1);
+		}
+#ifdef USE_PROCFS
+		if (proc_open(tcp, 0) < 0) {
+			fprintf(stderr, "trouble opening proc file\n");
+			cleanup();
+			exit(1);
+		}
+#endif /* USE_PROCFS */
+		break;
+	}
+}
+
 int
-main(argc, argv)
-int argc;
-char *argv[];
+main(int argc, char *argv[])
 {
 	extern int optind;
 	extern char *optarg;
 	struct tcb *tcp;
 	int c, pid = 0;
+	int optF = 0;
 	struct sigaction sa;
 
 	static char buf[BUFSIZ];
 
+	progname = argv[0] ? argv[0] : "strace";
+
 	/* Allocate the initial tcbtab.  */
 	tcbtabsize = argc;	/* Surely enough for all -p args.  */
-	tcbtab = (struct tcb **) malloc (tcbtabsize * sizeof tcbtab[0]);
-	tcbtab[0] = (struct tcb *) calloc (tcbtabsize, sizeof *tcbtab[0]);
+	if ((tcbtab = calloc (tcbtabsize, sizeof tcbtab[0])) == NULL) {
+		fprintf(stderr, "%s: out of memory\n", progname);
+		exit(1);
+	}
+	if ((tcbtab[0] = calloc (tcbtabsize, sizeof tcbtab[0][0])) == NULL) {
+		fprintf(stderr, "%s: out of memory\n", progname);
+		exit(1);
+	}
 	for (tcp = tcbtab[0]; tcp < &tcbtab[0][tcbtabsize]; ++tcp)
 		tcbtab[tcp - tcbtab[0]] = &tcbtab[0][tcp - tcbtab[0]];
 
-	progname = argv[0];
 	outf = stderr;
 	interactive = 1;
 	set_sortby(DEFAULT_SORTBY);
@@ -215,12 +660,12 @@ char *argv[];
 		case 'd':
 			debug++;
 			break;
+		case 'F':
+			optF = 1;
+			break;
 		case 'f':
 			followfork++;
 			break;
-		case 'F':
-			followvfork++;
-			break;
 		case 'h':
 			usage(stdout, 0);
 			break;
@@ -275,7 +720,7 @@ char *argv[];
 				fprintf(stderr, "%s: I'm sorry, I can't let you do that, Dave.\n", progname);
 				break;
 			}
-			if ((tcp = alloctcb(pid)) == NULL) {
+			if ((tcp = alloc_tcb(pid, 0)) == NULL) {
 				fprintf(stderr, "%s: out of memory\n",
 					progname);
 				exit(1);
@@ -311,9 +756,19 @@ char *argv[];
 		}
 	}
 
-	if (optind == argc && !pflag_seen)
+	if ((optind == argc) == !pflag_seen)
 		usage(stderr, 1);
 
+	if (!followfork)
+		followfork = optF;
+
+	if (followfork > 1 && cflag) {
+		fprintf(stderr,
+			"%s: -c and -ff are mutually exclusive options\n",
+			progname);
+		exit(1);
+	}
+
 	/* See if they want to run as another user. */
 	if (username != NULL) {
 		struct passwd *pent;
@@ -326,7 +781,7 @@ char *argv[];
 		}
 		if ((pent = getpwnam(username)) == NULL) {
 			fprintf(stderr, "%s: cannot find user `%s'\n",
-				progname, optarg);
+				progname, username);
 			exit(1);
 		}
 		run_uid = pent->pw_uid;
@@ -337,14 +792,8 @@ char *argv[];
 		run_gid = getgid();
 	}
 
-#ifndef SVR4
-	setreuid(geteuid(), getuid());
-#endif
-
 	/* Check if they want to redirect the output. */
 	if (outfname) {
-		long f;
-
 		/* See if they want to pipe the output. */
 		if (outfname[0] == '|' || outfname[0] == '!') {
 			/*
@@ -358,281 +807,34 @@ char *argv[];
 				exit(1);
 			}
 
-			if ((outf = popen(outfname + 1, "w")) == NULL) {
-				fprintf(stderr, "%s: can't popen '%s': %s\n",
-					progname, outfname + 1,
-					strerror(errno));
+			if ((outf = strace_popen(outfname + 1)) == NULL)
 				exit(1);
-			}
 		}
-		else if ((outf = fopen(outfname, "w")) == NULL) {
-			fprintf(stderr, "%s: can't fopen '%s': %s\n",
-				progname, outfname, strerror(errno));
+		else if (followfork <= 1 &&
+			 (outf = strace_fopen(outfname, "w")) == NULL)
 			exit(1);
-		}
-
-		if ((f=fcntl(fileno(outf), F_GETFD)) < 0 ) {
-			perror("failed to get flags for outputfile");
-			exit(1);
-		}
-
-		if (fcntl(fileno(outf), F_SETFD, f|FD_CLOEXEC) < 0 ) {
-			perror("failed to set flags for outputfile");
-			exit(1);
-		}
 	}
 
-#ifndef SVR4
-	setreuid(geteuid(), getuid());
-#endif
-
 	if (!outfname || outfname[0] == '|' || outfname[0] == '!')
 		setvbuf(outf, buf, _IOLBF, BUFSIZ);
 	if (outfname && optind < argc) {
 		interactive = 0;
 		qflag = 1;
 	}
+	/* Valid states here:
+	   optind < argc	pflag_seen	outfname	interactive
+	   1			0		0		1
+	   0			1		0		1
+	   1			0		1		0
+	   0			1		1		1
+	 */
 
-	for (c = 0; c < tcbtabsize; c++) {
-		tcp = tcbtab[c];
-		/* Reinitialize the output since it may have changed. */
-		tcp->outf = outf;
-		if (!(tcp->flags & TCB_INUSE) || !(tcp->flags & TCB_ATTACHED))
-			continue;
-#ifdef USE_PROCFS
-		if (proc_open(tcp, 1) < 0) {
-			fprintf(stderr, "trouble opening proc file\n");
-			droptcb(tcp);
-			continue;
-		}
-#else /* !USE_PROCFS */
-# ifdef LINUX
-		if (tcp->flags & TCB_CLONE_THREAD)
-			continue;
-		if (followfork) {
-			char procdir[MAXPATHLEN];
-			DIR *dir;
-			sprintf(procdir, "/proc/%d/task", tcp->pid);
-			dir = opendir(procdir);
-			if (dir != NULL) {
-				unsigned int ntid = 0, nerr = 0;
-				struct dirent *de;
-				int tid;
-				while ((de = readdir(dir)) != NULL) {
-					if (de->d_fileno == 0 ||
-					    de->d_name[0] == '.')
-						continue;
-					tid = atoi(de->d_name);
-					if (tid <= 0)
-						continue;
-					++ntid;
-					if (ptrace(PTRACE_ATTACH, tid,
-						   (char *) 1, 0) < 0)
-						++nerr;
-					else if (tid != tcbtab[c]->pid) {
-						if (nprocs == tcbtabsize &&
-						    expand_tcbtab())
-							tcp = NULL;
-						else
-							tcp = alloctcb(tid);
-						if (tcp == NULL) {
-							fprintf(stderr, "%s: out of memory\n",
-								progname);
-							exit(1);
-						}
-						tcp->flags |= TCB_ATTACHED|TCB_CLONE_THREAD|TCB_CLONE_DETACHED|TCB_FOLLOWFORK;
-						tcbtab[c]->nchildren++;
-						tcbtab[c]->nclone_threads++;
-						tcbtab[c]->nclone_detached++;
-						tcp->parent = tcbtab[c];
-					}
-				}
-				closedir(dir);
-				if (nerr == ntid) {
-					perror("attach: ptrace(PTRACE_ATTACH, ...)");
-					droptcb(tcp);
-					continue;
-				}
-				if (!qflag) {
-					ntid -= nerr;
-					if (ntid > 1)
-						fprintf(stderr, "\
-Process %u attached with %u threads - interrupt to quit\n",
-							tcp->pid, ntid);
-					else
-						fprintf(stderr, "\
-Process %u attached - interrupt to quit\n",
-							tcp->pid);
-				}
-				continue;
-			}
-		}
-# endif
-		if (ptrace(PTRACE_ATTACH, tcp->pid, (char *) 1, 0) < 0) {
-			perror("attach: ptrace(PTRACE_ATTACH, ...)");
-			droptcb(tcp);
-			continue;
-		}
-#endif /* !USE_PROCFS */
-		if (!qflag)
-			fprintf(stderr,
-				"Process %u attached - interrupt to quit\n",
-				tcp->pid);
-	}
-
-	if (!pflag_seen) {
-		struct stat statbuf;
-		char *filename;
-		char pathname[MAXPATHLEN];
-
-		filename = argv[optind];
-		if (strchr(filename, '/')) {
-			if (strlen(filename) > sizeof pathname - 1) {
-				errno = ENAMETOOLONG;
-				perror("strace: exec");
-				exit(1);
-			}
-			strcpy(pathname, filename);
-		}
-#ifdef USE_DEBUGGING_EXEC
-		/*
-		 * Debuggers customarily check the current directory
-		 * first regardless of the path but doing that gives
-		 * security geeks a panic attack.
-		 */
-		else if (stat(filename, &statbuf) == 0)
-			strcpy(pathname, filename);
-#endif /* USE_DEBUGGING_EXEC */
-		else {
-			char *path;
-			int m, n, len;
-
-			for (path = getenv("PATH"); path && *path; path += m) {
-				if (strchr(path, ':')) {
-					n = strchr(path, ':') - path;
-					m = n + 1;
-				}
-				else
-					m = n = strlen(path);
-				if (n == 0) {
-					getcwd(pathname, MAXPATHLEN);
-					len = strlen(pathname);
-				}
-				else if (n > sizeof pathname - 1)
-					continue;
-				else {
-					strncpy(pathname, path, n);
-					len = n;
-				}
-				if (len && pathname[len - 1] != '/')
-					pathname[len++] = '/';
-				strcpy(pathname + len, filename);
-				if (stat(pathname, &statbuf) == 0 &&
-				    /* Accept only regular files
-				       with some execute bits set.
-				       XXX not perfect, might still fail */
-				    S_ISREG(statbuf.st_mode) &&
-				    (statbuf.st_mode & 0111))
-					break;
-			}
-		}
-		if (stat(pathname, &statbuf) < 0) {
-			fprintf(stderr, "%s: %s: command not found\n",
-				progname, filename);
-			exit(1);
-		}
-		switch (pid = fork()) {
-		case -1:
-			perror("strace: fork");
-			cleanup();
-			exit(1);
-			break;
-		case 0: {
-#ifdef USE_PROCFS
-		        if (outf != stderr) close (fileno (outf));
-#ifdef MIPS
-			/* Kludge for SGI, see proc_open for details. */
-			sa.sa_handler = foobar;
-			sa.sa_flags = 0;
-			sigemptyset(&sa.sa_mask);
-			sigaction(SIGINT, &sa, NULL);
-#endif /* MIPS */
-#ifndef FREEBSD
-			pause();
-#else /* FREEBSD */
-			kill(getpid(), SIGSTOP); /* stop HERE */
-#endif /* FREEBSD */
-#else /* !USE_PROCFS */
-			if (outf!=stderr)
-				close(fileno (outf));
-
-			if (ptrace(PTRACE_TRACEME, 0, (char *) 1, 0) < 0) {
-				perror("strace: ptrace(PTRACE_TRACEME, ...)");
-				return -1;
-			}
-			if (debug)
-				kill(getpid(), SIGSTOP);
-
-			if (username != NULL || geteuid() == 0) {
-				uid_t run_euid = run_uid;
-				gid_t run_egid = run_gid;
-
-				if (statbuf.st_mode & S_ISUID)
-					run_euid = statbuf.st_uid;
-				if (statbuf.st_mode & S_ISGID)
-					run_egid = statbuf.st_gid;
-
-				/*
-				 * It is important to set groups before we
-				 * lose privileges on setuid.
-				 */
-				if (username != NULL) {
-					if (initgroups(username, run_gid) < 0) {
-						perror("initgroups");
-						exit(1);
-					}
-					if (setregid(run_gid, run_egid) < 0) {
-						perror("setregid");
-						exit(1);
-					}
-					if (setreuid(run_uid, run_euid) < 0) {
-						perror("setreuid");
-						exit(1);
-					}
-				}
-			}
-			else
-				setreuid(run_uid, run_uid);
-
-			/*
-			 * Induce an immediate stop so that the parent
-			 * will resume us with PTRACE_SYSCALL and display
-			 * this execve call normally.
-			 */
-			kill(getpid(), SIGSTOP);
-#endif /* !USE_PROCFS */
-
-			execv(pathname, &argv[optind]);
-			perror("strace: exec");
-			_exit(1);
-			break;
-		}
-		default:
-			if ((tcp = alloctcb(pid)) == NULL) {
-				fprintf(stderr, "tcb table full\n");
-				cleanup();
-				exit(1);
-			}
-#ifdef USE_PROCFS
-			if (proc_open(tcp, 0) < 0) {
-				fprintf(stderr, "trouble opening proc file\n");
-				cleanup();
-				exit(1);
-			}
-#endif /* USE_PROCFS */
-			break;
-		}
-	}
+	/* STARTUP_CHILD must be called before the signal handlers get
+	   installed below as they are inherited into the spawned process.
+	   Also we do not need to be protected by them as during interruption
+	   in the STARTUP_CHILD mode we kill the spawned process anyway.  */
+	if (!pflag_seen)
+		startup_child(&argv[optind]);
 
 	sigemptyset(&empty_set);
 	sigemptyset(&blocked_set);
@@ -671,37 +873,15 @@ Process %u attached - interrupt to quit\n",
 	sigaction(SIGCHLD, &sa, NULL);
 #endif /* USE_PROCFS */
 
+	if (pflag_seen)
+		startup_attach();
+
 	if (trace() < 0)
 		exit(1);
 	cleanup();
 	exit(0);
 }
 
-void
-newoutf(tcp)
-struct tcb *tcp;
-{
-	char name[MAXPATHLEN];
-	FILE *fp;
-
-	if (outfname && followfork > 1) {
-		sprintf(name, "%s.%u", outfname, tcp->pid);
-#ifndef SVR4
-		setreuid(geteuid(), getuid());
-#endif
-		fp = fopen(name, "w");
-#ifndef SVR4
-		setreuid(geteuid(), getuid());
-#endif
-		if (fp == NULL) {
-			perror("fopen");
-			return;
-		}
-		tcp->outf = fp;
-	}
-	return;
-}
-
 int
 expand_tcbtab()
 {
@@ -718,6 +898,8 @@ expand_tcbtab()
 	if (newtab == NULL || newtcbs == NULL) {
 		if (newtab != NULL)
 			free(newtab);
+		fprintf(stderr, "%s: expand_tcbtab: out of memory\n",
+			progname);
 		return 1;
 	}
 	for (i = tcbtabsize; i < 2 * tcbtabsize; ++i)
@@ -730,8 +912,7 @@ expand_tcbtab()
 
 
 struct tcb *
-alloctcb(pid)
-int pid;
+alloc_tcb(int pid, int command_options_parsed)
 {
 	int i;
 	struct tcb *tcp;
@@ -753,9 +934,12 @@ int pid;
 			tcp->stime.tv_usec = 0;
 			tcp->pfd = -1;
 			nprocs++;
+			if (command_options_parsed)
+				newoutf(tcp);
 			return tcp;
 		}
 	}
+	fprintf(stderr, "%s: alloc_tcb: tcb table full\n", progname);
 	return NULL;
 }
 
@@ -784,12 +968,7 @@ int attaching;
 		perror("strace: open(\"/proc/...\", ...)");
 		return -1;
 	}
-	if ((arg = fcntl(tcp->pfd, F_GETFD)) < 0) {
-		perror("F_GETFD");
-		return -1;
-	}
-	if (fcntl(tcp->pfd, F_SETFD, arg|FD_CLOEXEC) < 0) {
-		perror("F_SETFD");
+	if (set_cloexec_flag(tcp->pfd) < 0) {
 		return -1;
 	}
 	sprintf(proc, "/proc/%d/status", tcp->pid);
@@ -797,12 +976,7 @@ int attaching;
 		perror("strace: open(\"/proc/...\", ...)");
 		return -1;
 	}
-	if ((arg = fcntl(tcp->pfd_stat, F_GETFD)) < 0) {
-		perror("F_GETFD");
-		return -1;
-	}
-	if (fcntl(tcp->pfd_stat, F_SETFD, arg|FD_CLOEXEC) < 0) {
-		perror("F_SETFD");
+	if (set_cloexec_flag(tcp->pfd_stat) < 0) {
 		return -1;
 	}
 	sprintf(proc, "/proc/%d/as", tcp->pid);
@@ -810,12 +984,7 @@ int attaching;
 		perror("strace: open(\"/proc/...\", ...)");
 		return -1;
 	}
-	if ((arg = fcntl(tcp->pfd_as, F_GETFD)) < 0) {
-		perror("F_GETFD");
-		return -1;
-	}
-	if (fcntl(tcp->pfd_as, F_SETFD, arg|FD_CLOEXEC) < 0) {
-		perror("F_SETFD");
+	if (set_cloexec_flag(tcp->pfd_as) < 0) {
 		return -1;
 	}
 #else
@@ -830,12 +999,7 @@ int attaching;
 		perror("strace: open(\"/proc/...\", ...)");
 		return -1;
 	}
-	if ((arg = fcntl(tcp->pfd, F_GETFD)) < 0) {
-		perror("F_GETFD");
-		return -1;
-	}
-	if (fcntl(tcp->pfd, F_SETFD, arg|FD_CLOEXEC) < 0) {
-		perror("F_SETFD");
+	if (set_cloexec_flag(tcp->pfd) < 0) {
 		return -1;
 	}
 #endif
@@ -1137,6 +1301,16 @@ struct tcb *tcp;
 		if (!(tcp->flags & TCB_CLONE_DETACHED))
 #endif
 			tcp->parent->nzombies++;
+#ifdef LINUX
+		/* Update `tcp->parent->parent->nchildren' and the other fields
+		   like NCLONE_DETACHED, only for zombie group leader that has
+		   already reported and been short-circuited at the top of this
+		   function.  The same condition as at the top of DETACH.  */
+		if ((tcp->flags & TCB_CLONE_THREAD) &&
+		    tcp->parent->nclone_threads == 0 &&
+		    (tcp->parent->flags & TCB_EXITING))
+			droptcb(tcp->parent);
+#endif
 		tcp->parent = NULL;
 	}
 
@@ -1194,121 +1368,11 @@ struct tcb *tcp;
 	return 0;
 }
 
-#endif /* !USE_PROCFS */
-
-/* detach traced process; continue with sig */
-
 static int
-detach(tcp, sig)
-struct tcb *tcp;
-int sig;
+resume_from_tcp (struct tcb *tcp)
 {
 	int error = 0;
-#ifdef LINUX
-	int status, resumed;
-	struct tcb *zombie = NULL;
-
-	/* If the group leader is lingering only because of this other
-	   thread now dying, then detach the leader as well.  */
-	if ((tcp->flags & TCB_CLONE_THREAD) &&
-	    tcp->parent->nclone_threads == 1 &&
-	    (tcp->parent->flags & TCB_EXITING))
-		zombie = tcp->parent;
-#endif
-
-	if (tcp->flags & TCB_BPTSET)
-		sig = SIGKILL;
-
-#ifdef LINUX
-	/*
-	 * Linux wrongly insists the child be stopped
-	 * before detaching.  Arghh.  We go through hoops
-	 * to make a clean break of things.
-	 */
-#if defined(SPARC)
-#undef PTRACE_DETACH
-#define PTRACE_DETACH PTRACE_SUNDETACH
-#endif
-	if ((error = ptrace(PTRACE_DETACH, tcp->pid, (char *) 1, sig)) == 0) {
-		/* On a clear day, you can see forever. */
-	}
-	else if (errno != ESRCH) {
-		/* Shouldn't happen. */
-		perror("detach: ptrace(PTRACE_DETACH, ...)");
-	}
-	else if (kill(tcp->pid, 0) < 0) {
-		if (errno != ESRCH)
-			perror("detach: checking sanity");
-	}
-	else if (kill(tcp->pid, SIGSTOP) < 0) {
-		if (errno != ESRCH)
-			perror("detach: stopping child");
-	}
-	else {
-		for (;;) {
-#ifdef __WALL
-			if (wait4(tcp->pid, &status, __WALL, NULL) < 0) {
-				if (errno == ECHILD) /* Already gone.  */
-					break;
-				if (errno != EINVAL) {
-					perror("detach: waiting");
-					break;
-				}
-#endif /* __WALL */
-				/* No __WALL here.  */
-				if (waitpid(tcp->pid, &status, 0) < 0) {
-					if (errno != ECHILD) {
-						perror("detach: waiting");
-						break;
-					}
-#ifdef __WCLONE
-					/* If no processes, try clones.  */
-					if (wait4(tcp->pid, &status, __WCLONE,
-						  NULL) < 0) {
-						if (errno != ECHILD)
-							perror("detach: waiting");
-						break;
-					}
-#endif /* __WCLONE */
-				}
-#ifdef __WALL
-			}
-#endif
-			if (!WIFSTOPPED(status)) {
-				/* Au revoir, mon ami. */
-				break;
-			}
-			if (WSTOPSIG(status) == SIGSTOP) {
-				if ((error = ptrace(PTRACE_DETACH,
-				    tcp->pid, (char *) 1, sig)) < 0) {
-					if (errno != ESRCH)
-						perror("detach: ptrace(PTRACE_DETACH, ...)");
-					/* I died trying. */
-				}
-				break;
-			}
-			if ((error = ptrace(PTRACE_CONT, tcp->pid, (char *) 1,
-			    WSTOPSIG(status) == SIGTRAP ?
-			    0 : WSTOPSIG(status))) < 0) {
-				if (errno != ESRCH)
-					perror("detach: ptrace(PTRACE_CONT, ...)");
-				break;
-			}
-		}
-	}
-#endif /* LINUX */
-
-#if defined(SUNOS4)
-	/* PTRACE_DETACH won't respect `sig' argument, so we post it here. */
-	if (sig && kill(tcp->pid, sig) < 0)
-		perror("detach: kill");
-	sig = 0;
-	if ((error = ptrace(PTRACE_DETACH, tcp->pid, (char *) 1, sig)) < 0)
-		perror("detach: ptrace(PTRACE_DETACH, ...)");
-#endif /* SUNOS4 */
-
-#ifndef USE_PROCFS
-	resumed = 0;
+	int resumed = 0;
 
 	/* XXX This won't always be quite right (but it never was).
 	   A waiter with argument 0 or < -1 is waiting for any pid in
@@ -1364,18 +1428,152 @@ int sig;
 				}
 			}
 	}
+
+	return error;
+}
 #endif
 
 #endif /* !USE_PROCFS */
 
+/* detach traced process; continue with sig
+   Never call DETACH twice on the same process as both unattached and
+   attached-unstopped processes give the same ESRCH.  For unattached process we
+   would SIGSTOP it and wait for its SIGSTOP notification forever.  */
+
+static int
+detach(tcp, sig)
+struct tcb *tcp;
+int sig;
+{
+	int error = 0;
+#ifdef LINUX
+	int status, catch_sigstop;
+	struct tcb *zombie = NULL;
+
+	/* If the group leader is lingering only because of this other
+	   thread now dying, then detach the leader as well.  */
+	if ((tcp->flags & TCB_CLONE_THREAD) &&
+	    tcp->parent->nclone_threads == 1 &&
+	    (tcp->parent->flags & TCB_EXITING))
+		zombie = tcp->parent;
+#endif
+
+	if (tcp->flags & TCB_BPTSET)
+		sig = SIGKILL;
+
+#ifdef LINUX
+	/*
+	 * Linux wrongly insists the child be stopped
+	 * before detaching.  Arghh.  We go through hoops
+	 * to make a clean break of things.
+	 */
+#if defined(SPARC)
+#undef PTRACE_DETACH
+#define PTRACE_DETACH PTRACE_SUNDETACH
+#endif
+	/*
+	 * On TCB_STARTUP we did PTRACE_ATTACH but still did not get the
+	 * expected SIGSTOP.  We must catch exactly one as otherwise the
+	 * detached process would be left stopped (process state T).
+	 */
+	catch_sigstop = (tcp->flags & TCB_STARTUP);
+	if ((error = ptrace(PTRACE_DETACH, tcp->pid, (char *) 1, sig)) == 0) {
+		/* On a clear day, you can see forever. */
+	}
+	else if (errno != ESRCH) {
+		/* Shouldn't happen. */
+		perror("detach: ptrace(PTRACE_DETACH, ...)");
+	}
+	else if (my_tgkill((tcp->flags & TCB_CLONE_THREAD ? tcp->parent->pid
+							  : tcp->pid),
+			   tcp->pid, 0) < 0) {
+		if (errno != ESRCH)
+			perror("detach: checking sanity");
+	}
+	else if (!catch_sigstop && my_tgkill((tcp->flags & TCB_CLONE_THREAD
+					      ? tcp->parent->pid : tcp->pid),
+					     tcp->pid, SIGSTOP) < 0) {
+		if (errno != ESRCH)
+			perror("detach: stopping child");
+	}
+	else
+		catch_sigstop = 1;
+	if (catch_sigstop)
+		for (;;) {
+#ifdef __WALL
+			if (wait4(tcp->pid, &status, __WALL, NULL) < 0) {
+				if (errno == ECHILD) /* Already gone.  */
+					break;
+				if (errno != EINVAL) {
+					perror("detach: waiting");
+					break;
+				}
+#endif /* __WALL */
+				/* No __WALL here.  */
+				if (waitpid(tcp->pid, &status, 0) < 0) {
+					if (errno != ECHILD) {
+						perror("detach: waiting");
+						break;
+					}
+#ifdef __WCLONE
+					/* If no processes, try clones.  */
+					if (wait4(tcp->pid, &status, __WCLONE,
+						  NULL) < 0) {
+						if (errno != ECHILD)
+							perror("detach: waiting");
+						break;
+					}
+#endif /* __WCLONE */
+				}
+#ifdef __WALL
+			}
+#endif
+			if (!WIFSTOPPED(status)) {
+				/* Au revoir, mon ami. */
+				break;
+			}
+			if (WSTOPSIG(status) == SIGSTOP) {
+				if ((error = ptrace(PTRACE_DETACH,
+				    tcp->pid, (char *) 1, sig)) < 0) {
+					if (errno != ESRCH)
+						perror("detach: ptrace(PTRACE_DETACH, ...)");
+					/* I died trying. */
+				}
+				break;
+			}
+			if ((error = ptrace(PTRACE_CONT, tcp->pid, (char *) 1,
+			    WSTOPSIG(status) == SIGTRAP ?
+			    0 : WSTOPSIG(status))) < 0) {
+				if (errno != ESRCH)
+					perror("detach: ptrace(PTRACE_CONT, ...)");
+				break;
+			}
+		}
+#endif /* LINUX */
+
+#if defined(SUNOS4)
+	/* PTRACE_DETACH won't respect `sig' argument, so we post it here. */
+	if (sig && kill(tcp->pid, sig) < 0)
+		perror("detach: kill");
+	sig = 0;
+	if ((error = ptrace(PTRACE_DETACH, tcp->pid, (char *) 1, sig)) < 0)
+		perror("detach: ptrace(PTRACE_DETACH, ...)");
+#endif /* SUNOS4 */
+
+#ifndef USE_PROCFS
+	error |= resume_from_tcp (tcp);
+#endif
+
 	if (!qflag)
 		fprintf(stderr, "Process %u detached\n", tcp->pid);
 
 	droptcb(tcp);
 
 #ifdef LINUX
-	if (zombie != NULL)
-		error = detach(zombie) || error;
+	if (zombie != NULL) {
+		/* TCP no longer exists therefore you must not detach () it.  */
+		droptcb(zombie);
+	}
 #endif
 
 	return error;
@@ -1523,7 +1721,6 @@ rebuild_pollv()
 static void
 proc_poll_open()
 {
-	int arg;
 	int i;
 
 	if (pipe(proc_poll_pipe) < 0) {
@@ -1531,12 +1728,7 @@ proc_poll_open()
 		exit(1);
 	}
 	for (i = 0; i < 2; i++) {
-		if ((arg = fcntl(proc_poll_pipe[i], F_GETFD)) < 0) {
-			perror("F_GETFD");
-			exit(1);
-		}
-		if (fcntl(proc_poll_pipe[i], F_SETFD, arg|FD_CLOEXEC) < 0) {
-			perror("F_SETFD");
+		if (set_cloexec_flag(proc_poll_pipe[i]) < 0) {
 			exit(1);
 		}
 	}
@@ -1955,37 +2147,19 @@ handle_group_exit(struct tcb *tcp, int sig)
 			fprintf(stderr,
 				"PANIC: handle_group_exit: %d leader %d\n",
 				tcp->pid, leader ? leader->pid : -1);
-		detach(tcp);	/* Already died.  */
+		/* TCP no longer exists therefore you must not detach () it.  */
+#ifndef USE_PROCFS
+		resume_from_tcp (tcp);
+#endif
+		droptcb(tcp);	/* Already died.  */
 	}
 	else {
 		/* Mark that we are taking the process down.  */
 		tcp->flags |= TCB_EXITING | TCB_GROUP_EXITING;
 		if (tcp->flags & TCB_ATTACHED) {
-		  	if (leader != NULL && leader != tcp) {
-				if (leader->flags & TCB_ATTACHED) {
-					/* We need to detach the leader so
-					   that the process death will be
-					   reported to its real parent.
-					   But we kill it first to prevent
-					   it doing anything before we kill
-					   the whole process in a moment.
-					   We can use PTRACE_KILL on a
-					   thread that's not already
-					   stopped.  Then the value we pass
-					   in PTRACE_DETACH just sets the
-					   death signal reported to the
-					   real parent.  */
-					ptrace(PTRACE_KILL, leader->pid, 0, 0);
-					if (debug)
-						fprintf(stderr,
-							" [%d exit %d kills %d]\n",
-							tcp->pid, sig, leader->pid);
-					detach(leader, sig);
-				}
-				else
-					leader->flags |= TCB_GROUP_EXITING;
-			}
 			detach(tcp, sig);
+		  	if (leader != NULL && leader != tcp)
+				leader->flags |= TCB_GROUP_EXITING;
 		}
 		else if (ptrace(PTRACE_CONT, tcp->pid, (char *) 1, sig) < 0) {
 			perror("strace: ptrace(PTRACE_CONT, ...)");
@@ -2022,6 +2196,8 @@ trace()
 #endif /* LINUX */
 
 	while (nprocs != 0) {
+		if (interrupted)
+			return 0;
 		if (interactive)
 			sigprocmask(SIG_SETMASK, &empty_set, NULL);
 #ifdef LINUX
@@ -2054,9 +2230,6 @@ trace()
 		if (interactive)
 			sigprocmask(SIG_BLOCK, &blocked_set, NULL);
 
-		if (interrupted)
-			return 0;
-
 		if (pid == -1) {
 			switch (wait_errno) {
 			case EINTR:
@@ -2081,13 +2254,18 @@ trace()
 				return -1;
 			}
 		}
+		if (pid == popen_pid) {
+			if (WIFEXITED(status) || WIFSIGNALED(status))
+				popen_pid = -1;
+			continue;
+		}
 		if (debug)
 			fprintf(stderr, " [wait(%#x) = %u]\n", status, pid);
 
 		/* Look up `pid' in our table. */
 		if ((tcp = pid2tcb(pid)) == NULL) {
 #ifdef LINUX
-			if (followfork || followvfork) {
+			if (followfork) {
 				/* This is needed to go with the CLONE_PTRACE
 				   changes in process.c/util.c: we might see
 				   the child's initial trap before we see the
@@ -2097,13 +2275,16 @@ trace()
 				   will we have the association of parent and
 				   child so that we know how to do clearbpt
 				   in the child.  */
-				if ((tcp = alloctcb(pid)) == NULL) {
-					fprintf(stderr, " [tcb table full]\n");
+				if (nprocs == tcbtabsize &&
+				    expand_tcbtab())
+					tcp = NULL;
+				else
+					tcp = alloctcb(pid);
+				if (tcp == NULL) {
 					kill(pid, SIGKILL); /* XXX */
 					return 0;
 				}
 				tcp->flags |= TCB_ATTACHED | TCB_SUSPENDED;
-				newoutf(tcp);
 				if (!qflag)
 					fprintf(stderr, "\
 Process %d attached (waiting for parent)\n",
@@ -2166,6 +2347,7 @@ Process %d attached (waiting for parent)\n",
 #ifdef TCB_GROUP_EXITING
 			    && !(tcp->parent && (tcp->parent->flags &
 						 TCB_GROUP_EXITING))
+			    && !(tcp->flags & TCB_GROUP_EXITING)
 #endif
 				)
 				fprintf(stderr,
@@ -2194,35 +2376,24 @@ Process %d attached (waiting for parent)\n",
 			fprintf(stderr, "pid %u stopped, [%s]\n",
 				pid, signame(WSTOPSIG(status)));
 
-		if (tcp->flags & TCB_STARTUP) {
+		/*
+		 * Interestingly, the process may stop
+		 * with STOPSIG equal to some other signal
+		 * than SIGSTOP if we happend to attach
+		 * just before the process takes a signal.
+		 */
+		if ((tcp->flags & TCB_STARTUP) && WSTOPSIG(status) == SIGSTOP) {
 			/*
 			 * This flag is there to keep us in sync.
 			 * Next time this process stops it should
 			 * really be entering a system call.
 			 */
 			tcp->flags &= ~TCB_STARTUP;
-			if (tcp->flags & TCB_ATTACHED) {
-				/*
-				 * Interestingly, the process may stop
-				 * with STOPSIG equal to some other signal
-				 * than SIGSTOP if we happend to attach
-				 * just before the process takes a signal.
-				 */
-				if (!WIFSTOPPED(status)) {
-					fprintf(stderr,
-						"pid %u not stopped\n", pid);
-					detach(tcp, WSTOPSIG(status));
-					continue;
-				}
-			}
-			else {
-#ifdef SUNOS4
-				/* A child of us stopped at exec */
-				if (WSTOPSIG(status) == SIGTRAP && followvfork)
-					fixvfork(tcp);
-#endif /* SUNOS4 */
-			}
 			if (tcp->flags & TCB_BPTSET) {
+				/*
+				 * One example is a breakpoint inherited from
+				 * parent through fork ().
+				 */
 				if (clearbpt(tcp) < 0) /* Pretty fatal */ {
 					droptcb(tcp);
 					cleanup();
@@ -2250,11 +2421,12 @@ Process %d attached (waiting for parent)\n",
 			}
 			if (!cflag
 			    && (qual_flags[WSTOPSIG(status)] & QUAL_SIGNAL)) {
-				unsigned long addr = 0, pc = 0;
-#ifdef PT_GETSIGINFO
+				unsigned long addr = 0;
+				long pc = 0;
+#if defined(PT_CR_IPSR) && defined(PT_CR_IIP) && defined(PT_GETSIGINFO)
 #				define PSR_RI	41
 				struct siginfo si;
-				unsigned long psr;
+				long psr;
 
 				upeek(pid, PT_CR_IPSR, &psr);
 				upeek(pid, PT_CR_IIP, &pc);
@@ -2297,6 +2469,9 @@ Process %d attached (waiting for parent)\n",
 			tcp->flags &= ~TCB_SUSPENDED;
 			continue;
 		}
+		/* we handled the STATUS, we are permitted to interrupt now. */
+		if (interrupted)
+			return 0;
 		if (trace_syscall(tcp) < 0) {
 			if (tcp->flags & TCB_ATTACHED)
 				detach(tcp, 0);
